//-------+---------+---------+---------+---------+---------+---------+--------=
//
// File: segment.c
//
//----------
//
// segment--
//	Support for lists of matched segments of DNA sequences.
//
// This module implements the collection of a table of segments of a pair of
// DNA sequences.  A segment consists of an interval in each sequence (of the
// same length) and a score.  The actual DNA content of the interval is of
// no importance to this module-- all that it important abotu it is represented
// by a segment's length and score.
//
// The caller can specify a limit on the total length of segments in the table.
// The intent is to give the user the ability to collect alignments covering
// some meaningful portion of the sequences, even though she does not know what
// score threshold would accomplish this.  Without this capability, the user is
// forced to do some preliminary runs to try to estimate the score distribution,
// or set the score threshold so low that the process is overwhelmed by false
// positives.
//
// The limit (coverageLimit) is the number of bases of coverage that the caller
// wishes to allow.  However, we allow the sum of all the segment lengths to
// exceed this, by no more than one segment.  For example, if the caller
// specifies a limit of 10K bp, and if there is at least 10K bp of homologous
// segments in the two sequences, we will keep at least 10K bp in this table,
// but not much more.
//
//----------
//
// $$$ Caveats
//
// The coverageLimit stuff doesn't consider the possibility that the incoming
// segments could contain overlaps.  As of this writing (Mar/2008) we will only
// have overlaps if the caller is using a twin-hit seed.
//
//----------

//----------
//
// other files
//
//----------

#include <stdlib.h>				// standard C stuff
#define  true  1
#define  false 0
#include "dna_utilities.h"		// dna/scoring stuff
#include "sequences.h"			// sequence stuff
#include "diag_hash.h"			// diagonals hashing stuff

#define  segment_owner			// (make this the owner of its globals)
#include "segment.h"			// interface to this module

// miscellany

#define segtable_bytes(size) (sizeof(segtable) + (((size)-1)*sizeof(segment)))

//#define debugBinaryHeap

//----------
//
// prototypes for private functions
//
//----------

static void remove_root       (segtable* st);
static void record_tie_scores (segtable* st);
static int  record_tie_score  (segtable* st, int ix);

#ifdef debugBinaryHeap
static void validate_heap     (segtable* st, char* msg);
#endif // debugBinaryHeap

//----------
//
// new_segment_table--
//	Allocate am empty segment table.
//
//----------
//
// Arguments:
//	u32		size:			The number of entries to provide for.  This can be
//							.. increased later.
//	unspos	coverageLimit:	Limit on the total lengths of the segments in the
//							.. table (see discussion in file header above);
//							.. zero indicates no limit.
//
// Returns:
//	A pointer to a newly allocated segment table;  failures result in program
//	fatality.  The caller must eventually dispose of the table, with a call to
//	free_segment_table().
//
//----------

segtable* new_segment_table
   (u32			size,
	unspos		coverageLimit)
	{
	segtable*	st;
	u32			bytesNeeded;

	// sanity check

	if (size < 1)
		suicidef ("in new_segment_table(), size can't be %d", size);

	// allocate

	bytesNeeded = segtable_bytes (size);
	st = (segtable*) malloc_or_die ("new_segment_table", bytesNeeded);

	// initialize

	st->size          = size;
	st->len           = 0;
	st->coverageLimit = coverageLimit;
	st->coverage      = 0;
	st->lowScore      = worstPossibleScore;

	return st;
	}

//----------
//
// empty_segment_table--
//	"Erase" all segments from a segment table.
//
//----------
//
// Arguments:
//	segtable*	st:	The segment table to empty.
//
// Returns:
//	(nothing)
//
//----------

void empty_segment_table
   (segtable*	st)
	{
	st->len      = 0;
	st->coverage = 0;
	st->lowScore = worstPossibleScore;
	}

//----------
//
// limit_segment_table--
//	Change a segment table's total length limit.
//
//----------
//
// Arguments:
//	segtable*	st:				The segment table to modify.
//	unspos		coverageLimit:	Limit on the total lengths of the segments in
//								.. the table;  zero indicates no limit.
//
// Returns:
//	(nothing)
//
//----------

void limit_segment_table
   (segtable*	st,
	unspos		coverageLimit)
	{
	int			ix, newLen;
	possum		cov, newCov;
	score		prevScore, newLow;
	segment*	seg, *tail;
	segment		tempSeg;

	st->coverageLimit = coverageLimit;

	// if this leaves us below the limit, we're done

	if ((st->coverageLimit == 0)
	 || (st->coverage < st->coverageLimit)
	 || (st->len < 2))
	 	return;

	// otherwise, may need to reduce;  begin by sorting segments by decreasing
	// score

	sort_segments (st, qSegmentsByDecreasingScore);

	// find the score that reduces coverage to no more than the limit

	newLen = st->len;
	newCov = st->coverage;
	newLow = st->lowScore;

	ix = st->len;
	seg = &st->seg[--ix];

	cov       = st->coverage - seg->length;
	prevScore = seg->s;

	while (ix > 0)
		{
		seg = &st->seg[--ix];

		// if this is the same as the segment after it, just keep looking (we
		// treat segments with the same score as one unbreakable entity)

		if (seg->s == prevScore)
			{ cov -= seg->length;  continue; }

		// if removing the segments above this one would be enough to get us
		// within the limit, we are finished;  note that what we remove will be
		// the segments above the (score-tied) segments above this one

		if (cov < st->coverageLimit)
			break;

		// potentially, we will remove the segments above this one, so record
		// that information and keep going

		newLen = ix+1;
		newCov = cov;
		newLow = seg->s;

		cov -= seg->length;
		prevScore = seg->s;
		}

	st->len      = newLen;
	st->coverage = newCov;
	st->lowScore = newLow;

	// now make the list a proper min-heap by simply reversing its order (which
	// orders it by increasing score), and add the tied-score information

	seg  = &st->seg[0];
	tail = &st->seg[st->len-1];
	while (seg < tail)
		{
		tempSeg   = *seg;
		*(seg++)  = *tail;
		*(tail--) = tempSeg;
		}

	record_tie_scores (st);
	}

//----------
//
// free_segment_table--
//	Deallocate a segment table.
//
//----------
//
// Arguments:
//	segtable*	st:	The segment table to dispose of.
//
// Returns:
//	(nothing)
//
//----------

void free_segment_table
   (segtable*	st)
	{
	free_if_valid ("free_segment_table", st);
	}

//----------
//
// add_segment--
//	Add a segment to a segment table.
//
//----------
//
// Arguments:
//	segtable*	st:				The segment table to add to.
//	pos1, pos2, length, s, id:	The segment to add.
//
// Returns:
//	A pointer to a the segment table;  if there was room in st for the segment,
//	this is the same as st;  otherwise, this is a pointer to a new copy of st,
//	and the previous st has been deallocated;  reallocation failures result in
//	program fatality.
//
//----------

segtable* add_segment
   (segtable*	st,
	unspos		pos1,
	unspos		pos2,
	unspos		length,
	score		s,
	int			id)
	{
	u32			newSize;
	u32			bytesNeeded;
	segment*	seg, *parent;
	segment		tempSeg;
	int			ix, pIx;
	int			tied, stopped;

	//fprintf (stderr, "add " unsposSlashSFmt " " unsposFmt " " scoreFmtSimple "; id %d\n",
	//               pos1+1, "+",
	//               pos2+1, ((id & rcf_rev) != 0)? "-" : "+",
	//               length, s, id);

	//////////
	// add the segment to the table, enlarging the table if needed, but
	// discarding the segment if it is low-scoring and the table has met its
	// coverage limit
	//////////

	// if the table is already full and this segment scores less than the
	// lowest score in the table, discard it

	if ((st->len > 0)
	 && (st->coverageLimit != 0)
	 && (st->coverage >= st->coverageLimit)
	 && (s < st->lowScore))
		return st;

	// if there's no room for the new segment, re-allocate

	if (st->len >= st->size)
		{
		newSize     = st->size + 100 + (st->size / 3);
		bytesNeeded = segtable_bytes (newSize);
		st = (segtable*) realloc_or_die ("add_segment", st, bytesNeeded);
		st->size = newSize;
		}

	// add the segment, by appending it at the end

	seg = &st->seg[st->len++];
	seg->pos1     = pos1;
	seg->pos2     = pos2;
	seg->length   = length;
	seg->s        = s;
	seg->id       = id;
	seg->filter   = false;
	seg->scoreCov = (possum) length;

	st->coverage += length;
	if ((st->len == 1) || (s < st->lowScore)) st->lowScore = s;

	//////////
	// handle the transition between the two table states
	//	below-the-coverage-limit:  table is kept as a simple list
	//	met-the-coverage-limit:    table is kept as a proper min-heap
	//////////

	// if this segment leaves us below the limit, we're done

	if ((st->coverageLimit == 0)
	 || (st->coverage < st->coverageLimit))
		return st;

	// if this is the first time we've reached the limit, sort the segments to
	// create a proper min-heap, and add the tied-score information
	// nota bene:  if we reach here, st->coverageLimit > 0 and
	//             st->coverage >= st->coverageLimit

	if (st->coverage - length < st->coverageLimit)
		{
		sort_segments (st, qSegmentsByIncreasingScore);
		record_tie_scores (st);
		#ifdef debugBinaryHeap
		fprintf       (stderr, "\nafter sort:\n");
		dump_segments (stderr, st, NULL, NULL);
		validate_heap (st, "after sort");
		#endif // debugBinaryHeap
		goto prune;
		}

	//////////
	// maintain the min-heap property
	//////////

	#ifdef debugBinaryHeap
	//fprintf       (stderr, "\nbefore percolation:\n");
	//dump_segments (stderr, st, NULL, NULL);
	#endif // debugBinaryHeap

	// the rest of the list is a proper min-heap, so percolate the new segment
	// up the tree, while maintaining the tied-score information
	// nota bene:  if we reach here, length >= 2

	tied = false;
	for (ix=st->len-1 ; ix>0 ; )
		{
		pIx    = (ix-1) / 2;
		seg    = &st->seg[ix];
		parent = &st->seg[pIx];

		if (seg->s >= parent->s)
			{ tied = (seg->s == parent->s);  break; }

		// swap this segment with its parent, and adjust old parent's tied-score
		// subheap

		tempSeg = *seg;  *seg = *parent;  *parent = tempSeg;
		record_tie_score (st, ix);

		ix = pIx;
		}

	record_tie_score (st, ix);

	// if the new segment tied an existing score, we must continue to percolate
	// the tied-score info up the tree

	if (tied)
		{
		stopped = false;
		for (ix=(ix-1)/2 ; ix>0 ; ix=(ix-1)/2)
			{
			if (!record_tie_score (st, ix))
				{ stopped = true;  break; }
			}
		if (!stopped) record_tie_score (st, 0);
		}

	#ifdef debugBinaryHeap
	fprintf       (stderr, "\nafter percolation:\n");
	dump_segments (stderr, st, NULL, NULL);
	validate_heap (st, "after percolation");
	#endif // debugBinaryHeap

	//////////
	// remove low-scoring segments
	//////////

prune:

	// if removing the minimum scoring subheap would bring us below the
	// limit, no pruning is necessary

	if (st->coverage - st->seg[0].scoreCov < st->coverageLimit)
		return st;

	// otherwise, we must remove subheaps as long as doing so leaves us at or
	// above the limit

	while (st->coverage - st->seg[0].scoreCov >= st->coverageLimit)
		{
		s = st->seg[0].s;
		while (st->seg[0].s == s)
			{
			remove_root (st);
			#ifdef debugBinaryHeap
			fprintf       (stderr, "\nafter a pruning:\n");
			dump_segments (stderr, st, NULL, NULL);
			validate_heap (st, "after pruning");
			#endif // debugBinaryHeap
			}
		}

	st->lowScore = st->seg[0].s;

	#ifdef debugBinaryHeap
	fprintf       (stderr, "\nafter pruning:\n");
	dump_segments (stderr, st, NULL, NULL);
	validate_heap (st, "after pruning");
	#endif // debugBinaryHeap

	return st;
	}


static void remove_root (segtable* st)
	{
	segment*	seg, *detached, *child, *rgtChild;
	u32			ix, childIx, rgtIx;

	// remove the root node's coverage

	seg = &st->seg[0];
	st->coverage -= seg->length;

	if (st->len <= 1)
		{ st->len = 0;  return; }

	// detach the final segment;  conceptually we will consider this as the new
	// root, then percolate it down to a new position

	detached = &st->seg[--st->len];

	if (st->len == 1)
		{ *seg = *detached;  return; }

	// recompute tied-score info up the tree from the detached node
	// nota bene:  st->len > 1, so the initial value of ix, (st->len-1)/2 >= 0

	for (ix=(st->len-1)/2 ; ix>0 ; ix=(ix-1)/2)
		{ if (!record_tie_score (st, ix)) break; }

	// the detached node may violate the min-heap property;  percolate it down
	// the tree until the property is re-established

	ix = 0;
	while (true)
		{
		childIx = 2*ix+1;
		if (childIx >= st->len) break;	// (ix is a leaf position)
		child = &st->seg[childIx];

		rgtIx = childIx + 1;
		if (rgtIx < st->len)
			{
			rgtChild = &st->seg[rgtIx];
			if (rgtChild->s < child->s)
				{ childIx = rgtIx;  child = rgtChild; }
			}

		if (detached->s <= child->s)// (if we place the detached node here
			break;					//  .. we'll satisfy the min-heap property)

		st->seg[ix] = *child;		// move the child up the tree

		ix = childIx;
		}

	st->seg[ix] = *detached;		// copy the detached node into the tree

	// now reverse that path, up the tree, to recompute tied-score info

	for ( ; ix>0 ; ix=(ix-1)/2)
		record_tie_score (st, ix);
	record_tie_score (st, 0);
	}

//----------
//
// record_tie_scores--
//	Record tied-score information in a segment table.
//
//----------
//
// Arguments:
//	segtable*	st:		The segment table.
//
// Returns:
//	(nothing)
//
//----------

static void record_tie_scores
   (segtable*	st)
	{
	int			ix;

	// determine the coverage of all equal-scoring subheaps;  we start at the
	// tail and percolate coverage up the tree whenever we have a tie;  at the
	// end, seg->scoreCov will be equal to the sum of the lengths of the
	// subheap rooted at seg that consists of segments having the same score

	for (ix=st->len-1 ; ix>=0 ; ix--)
		record_tie_score (st, ix);
	}

//----------
//
// record_tie_score--
//	Record tied-score information for one node in a segment table.
//
//----------
//
// Arguments:
//	segtable*	st:		The segment table.
//	int			ix:		The node index.
//
// Returns:
//	true  => score coverage for the node was changed by this operation;
//	false => score coverage for the node remained the same
//
//----------

static int record_tie_score
   (segtable*	st,
	int			ix)
	{
	possum		scoreCov;
	segment*	seg, *lftChild, *rgtChild;
	u32			lftIx, rgtIx;

	seg = &st->seg[ix];

	// scoreCov is the sum of the length of this segment, plus the left and
	// right subheaps *if* they exist and have the same score as this segment

	scoreCov = (possum) seg->length;

	lftIx = 2*ix+1;
	if (lftIx < st->len)
		{
		lftChild = &st->seg[lftIx];
		if (lftChild->s == seg->s)
			scoreCov += lftChild->scoreCov;

		rgtIx = lftIx + 1;
		if (rgtIx < st->len)
			{
			rgtChild = &st->seg[rgtIx];
			if (rgtChild->s == seg->s)
				scoreCov += rgtChild->scoreCov;
			}
		}

	if (scoreCov != seg->scoreCov)
		{ seg->scoreCov = scoreCov;  return true;  }

	return false;
	}

//----------
//
// split_segment_table--
//	Split a segment table into two tables.  All segments with a particular id
//	are left in the table, and all others are moved to a new table.
//
//----------
//
// Arguments:
//	segtable*	st:			The segment table to split.
//	int			id:			The id of segments to keep in the incoming table.
//	segtable**	leftovers:	The segment table in which to collect the leftovers.
//							.. This table must already exist.
//
// Returns:
//	(nothing)
//
//----------

void split_segment_table
   (segtable*	st,
	int			id,
	segtable**	_leftovers)
	{
	segtable*	leftovers = *_leftovers;
	possum		cov;
	score		lowScore;
	u32			srcIx, dstIx;
	segment*	seg;

	cov      = 0;
	lowScore = worstPossibleScore;

	//fprintf       (stderr, "incoming\n");
	//dump_segments (stderr, st, NULL, NULL);

	dstIx = 0;
	for (srcIx=0 ; srcIx<st->len ; srcIx++)
		{
		seg = &st->seg[srcIx];

		if (seg->id != id)
			{
			leftovers = add_segment (leftovers,
			                         seg->pos1, seg->pos2, seg->length,
			                         seg->s, seg->id);
			continue;
			}

		cov += seg->length;
		if ((dstIx == 0) || (seg->s < lowScore)) lowScore = seg->s;

		if (dstIx != srcIx) st->seg[dstIx] = *seg;
		dstIx++;
		}

	st->len      = dstIx;
	st->coverage = cov;
	st->lowScore = lowScore;

	//fprintf       (stderr, "\nid=%d\n", id);
	//dump_segments (stderr, st,        NULL, NULL);
	//fprintf       (stderr, "\nileftovers\n");
	//dump_segments (stderr, leftovers, NULL, NULL);
	//fprintf       (stderr, "\n");

	*_leftovers = leftovers;
	}

//----------
//
// sort_segments--
//	Sort a segment table.
//
//----------
//
// Arguments:
//	segtable*	st:		The segment table to add to.
//	int (*qCompare):	Comparison function to use (suitable for the standard
//						.. c function qsort).  For example, this could be
//						.. qSegmentsByPos2.
//
// Returns:
//	(nothing)
//
//----------

void sort_segments
   (segtable*	st,
	int (*qCompare) (const void* el1, const void* el2))
	{
	qsort (st->seg, st->len, sizeof(segment), qCompare);
	}

//----------
//
// merge_segments--
//	Merge any overlapping segments in a segment table.
//
//----------
//
// Arguments:
//	segtable*	st:	The segment table to operate upon.
//
// Returns:
//	(nothing)
//
//----------
//
// Notes:
//	(1)	Segments are considered to overlap if they are on the same diagonal and
//		share any positions.  Segments that adjoin, but do not overlap, are not
//		merged.
//	(2)	Any merged segment is given the maximum score of the segments being
//		merged.  If this is not appropriate, the caller will need to rescore
//		the segments afterwards.
//		$$$ The problem with making the caller rescore is that the caller has
//		$$$ .. no way to know which segments have been merged, thus *every*
//		$$$ .. segment would have to be rescored.  We could resolve this by
//		$$$ .. having the caller provide a callback routine to rescore a
//		$$$ .. segment.
//	(3) The min-heap property is NOT maintained.  We view this routine as a
//		post-processing step, and we do not expect additional segments to be
//		added to the table after this is called.
//
//----------

void merge_segments
   (segtable*	st)
	{
	u32			srcIx,   dstIx;
	segment*	srcSeg, *dstSeg;
	unspos		pos2, end2, srcPos2, srcEnd2;
	sgnpos		diag,       srcDiag;
	score		s,          srcS;

	// if we have fewer than two segments, there's nothing to merge

	if (st->len < 2) return;

	// sort segments by increasing pos2 along each diagonal

	sort_segments (st, qSegmentsByDiag);

	// start first segment

	srcSeg = st->seg;
	pos2   = srcSeg->pos2;
	diag   = diagNumber (srcSeg->pos1, pos2);
	end2   = pos2 + srcSeg->length;
	s      = srcSeg->s;
	srcSeg++;

	// scan segments, merging as needed;  note that any segment written to the
	// list is written to an earlier (or the same) index as the latest read

	dstIx = 0;
	for (srcIx=1 ; srcIx<st->len ; srcIx++,srcSeg++)
		{
		srcPos2 = srcSeg->pos2;
		srcDiag = diagNumber (srcSeg->pos1, srcPos2);
		srcEnd2 = srcPos2 + srcSeg->length;
		srcS    = srcSeg->s;

		if ((srcDiag == diag) && (srcPos2 < end2))
			{ // merge
			if (srcEnd2 > end2) end2 = srcEnd2;
			if (srcS    > s)    s    = srcS;
			continue;
			}

		// deposit the previous segment

		dstSeg = &st->seg[dstIx++];
		dstSeg->pos1   = (unspos) (diag + pos2);
		dstSeg->pos2   = pos2;
		dstSeg->length = end2 - pos2;
		dstSeg->s      = s;

		// start a new segment

		pos2 = srcPos2;
		diag = srcDiag;
		end2 = srcEnd2;
		s    = srcS;
		}

	// deposit the final segment

	dstSeg = &st->seg[dstIx++];
	dstSeg->pos1   = (unspos) (diag + pos2);
	dstSeg->pos2   = pos2;
	dstSeg->length = end2 - pos2;
	dstSeg->s      = s;

	// adjust the length of the list

	st->len = dstIx;
	}

//----------
//
// filter_marked_segments--
//	Remove from a table any segments marked for filtering.
//
//----------
//
// Arguments:
//	segtable*	st:	The segment table to operate upon.
//
// Returns:
//	(nothing)
//
//----------

void filter_marked_segments
   (segtable*	st)
	{
	segment*	srcSeg, *dstSeg;

	if (st      == NULL) return;
	if (st->seg == NULL) return;

	for (dstSeg=srcSeg=st->seg ; ((u32)(srcSeg-st->seg))<st->len ; srcSeg++)
		{
		if (srcSeg->filter) continue;
		if (srcSeg != dstSeg) *dstSeg = *srcSeg;
		dstSeg++;
		}

	st->len = dstSeg - st->seg;
	}

//----------
// [[-- comparison function for the standard c function qsort --]]
//
// qSegmentsByPos1--
//	Compare two segments by pos1, so that qsort will sort by increasing pos1.
//
//----------
//
// Arguments:
//	const void* (really segment*) segA:	Pointer to one segment.
//	const void* (really segment*) segB:	Pointer to another.
//
// Returns:
//	> 0 => segA is greater than segB.
//	= 0 => segA and segB are the same.
//	< 0 => segA is less than segB.
//
//----------

int qSegmentsByPos1
   (const void*	_segA,
	const void*	_segB)
	{
	segment*	segA = (segment*) _segA;
	segment*	segB = (segment*) _segB;

	if      (segA->pos1   < segB->pos1)   return -1;
	else if (segA->pos1   > segB->pos1)   return  1;

	if      (segA->length < segB->length) return -1;
	else if (segA->length > segB->length) return  1;

	if      (segA->pos2   < segB->pos2)   return -1;
	else if (segA->pos2   > segB->pos2)   return  1;

	if      (segA->id     < segB->id)     return -1;
	else if (segA->id     > segB->id)     return  1;

	if      (segA->s      < segB->s)      return -1;
	else if (segA->s      > segB->s)      return  1;

	return 0;
	}

//----------
// [[-- comparison function for the standard c function qsort --]]
//
// qSegmentsByPos2--
//	Compare two segments by pos2, so that qsort will sort by increasing pos2.
//
//----------
//
// Arguments:
//	const void* (really segment*) segA:	Pointer to one segment.
//	const void* (really segment*) segB:	Pointer to another.
//
// Returns:
//	> 0 => segA is greater than segB.
//	= 0 => segA and segB are the same.
//	< 0 => segA is less than segB.
//
//----------

int qSegmentsByPos2
   (const void*	_segA,
	const void*	_segB)
	{
	segment*	segA = (segment*) _segA;
	segment*	segB = (segment*) _segB;

	if      (segA->pos2   < segB->pos2)   return -1;
	else if (segA->pos2   > segB->pos2)   return  1;

	if      (segA->length < segB->length) return -1;
	else if (segA->length > segB->length) return  1;

	if      (segA->pos1   < segB->pos1)   return -1;
	else if (segA->pos1   > segB->pos1)   return  1;

	if      (segA->id     < segB->id)     return -1;
	else if (segA->id     > segB->id)     return  1;

	if      (segA->s      < segB->s)      return -1;
	else if (segA->s      > segB->s)      return  1;

	return 0;
	}

//----------
// [[-- comparison function for the standard c function qsort --]]
//
// qSegmentsByDecreasingScore--
//	Compare two segments by score, so that qsort will sort by decreasing score.
// qSegmentsByIncreasingScore--
//	Compare two segments by score, so that qsort will sort by increasing score.
//
//----------
//
// Arguments:
//	const void* (really segment*) segA:	Pointer to one segment.
//	const void* (really segment*) segB:	Pointer to another.
//
// Returns:
//	(for qSegmentsByDecreasingScore)     (for qSegmentsByIncreasingScore)
//	> 0 => segA is greater than segB.    > 0 => segA is greater than segB.
//	= 0 => segA and segB are the same.   = 0 => segA and segB are the same.
//	< 0 => segA is less than segB.       < 0 => segA is less than segB.
//
//----------

int qSegmentsByDecreasingScore
   (const void*	_segA,
	const void*	_segB)
	{
	segment*	segA = (segment*) _segA;
	segment*	segB = (segment*) _segB;

	if      (segA->s      < segB->s)      return  1;
	else if (segA->s      > segB->s)      return -1;

	if      (segA->length < segB->length) return -1; // if scores are equal we
	else if (segA->length > segB->length) return  1; // .. prefer shorter length

	if      (segA->pos2   < segB->pos2)   return -1;
	else if (segA->pos2   > segB->pos2)   return  1;

	if      (segA->pos1   < segB->pos1)   return -1;
	else if (segA->pos1   > segB->pos1)   return  1;

	if      (segA->id     < segB->id)     return -1;
	else if (segA->id     > segB->id)     return  1;

	return 0;
	}


int qSegmentsByIncreasingScore
   (const void*	_segA,
	const void*	_segB)
	{
	segment*	segA = (segment*) _segA;
	segment*	segB = (segment*) _segB;

	if      (segA->s      < segB->s)      return -1;
	else if (segA->s      > segB->s)      return  1;

	if      (segA->length < segB->length) return -1; // if scores are equal we
	else if (segA->length > segB->length) return  1; // .. prefer shorter length

	if      (segA->pos2   < segB->pos2)   return -1;
	else if (segA->pos2   > segB->pos2)   return  1;

	if      (segA->pos1   < segB->pos1)   return -1;
	else if (segA->pos1   > segB->pos1)   return  1;

	if      (segA->id     < segB->id)     return -1;
	else if (segA->id     > segB->id)     return  1;

	return 0;
	}

//----------
// [[-- comparison function for the standard c function qsort --]]
//
// qSegmentsByDiag--
//	Compare two segments by diagonal, so that qsort will sort by increasing
//	diagonal, and by increasing pos2 along each diagonal. 
//
//----------
//
// Arguments:
//	const void* (really segment*) segA:	Pointer to one segment.
//	const void* (really segment*) segB:	Pointer to another.
//
// Returns:
//	> 0 => segA is greater than segB.
//	= 0 => segA and segB are the same.
//	< 0 => segA is less than segB.
//
//----------

int qSegmentsByDiag
   (const void*	_segA,
	const void*	_segB)
	{
	segment*	segA = (segment*) _segA;
	segment*	segB = (segment*) _segB;
	sgnpos		diagA, diagB;

	// compare by diagonal

	diagA = diagNumber (segA->pos1, segA->pos2);
	diagB = diagNumber (segB->pos1, segB->pos2);

	if      (diagA < diagB) return -1;
	else if (diagA > diagB) return  1;

	// resort to tiebreakers

	if      (segA->pos2   < segB->pos2)   return -1;
	else if (segA->pos2   > segB->pos2)   return  1;

	if      (segA->length < segB->length) return -1;
	else if (segA->length > segB->length) return  1;

	if      (segA->id     < segB->id)     return -1;
	else if (segA->id     > segB->id)     return  1;

	if      (segA->s      < segB->s)      return -1;
	else if (segA->s      > segB->s)      return  1;

	return 0;
	}

//----------
// [[-- comparison function for the standard c function qsort --]]
//
// qSegmentsById--
//	Compare two segments by id, so that qsort will sort by increasing id.
//
//----------
//
// Arguments:
//	const void* (really segment*) segA:	Pointer to one segment.
//	const void* (really segment*) segB:	Pointer to another.
//
// Returns:
//	> 0 => segA is greater than segB.
//	= 0 => segA and segB are the same.
//	< 0 => segA is less than segB.
//
//----------

int qSegmentsById
   (const void*	_segA,
	const void*	_segB)
	{
	segment*	segA = (segment*) _segA;
	segment*	segB = (segment*) _segB;

	if      (segA->id     < segB->id)     return -1;
	else if (segA->id     > segB->id)     return  1;

	if      (segA->s      < segB->s)      return -1;
	else if (segA->s      > segB->s)      return  1;

	if      (segA->pos1   < segB->pos1)   return -1;
	else if (segA->pos1   > segB->pos1)   return  1;

	if      (segA->length < segB->length) return -1;
	else if (segA->length > segB->length) return  1;

	if      (segA->pos2   < segB->pos2)   return -1;
	else if (segA->pos2   > segB->pos2)   return  1;

	return 0;
	}

//----------
//
// dump_segments--
//	Dump a segment table, for debugging.
//
//----------
//
// Arguments:
//	FILE*		f:			The file to print to.
//	segtable*	st:			The segment table to dump.
//	char*		sym1,sym2:	Identifying strings to attach to pos1 and pos2,
//							.. respectively.  E.g. "+" or "-" for DNA strand.
//							.. These can be NULL, in which case we presume that
//							.. segment id fields are the second sequence's
//							.. revCompFlags, and we take the strand from that.
//
// Returns:
//	(nothing)
//
//----------

//#define show_separation

void dump_segments
   (FILE*		f,
	segtable*	st,
	char*		_sym1,
	char*		_sym2)
	{
	char*		sym1, *sym2;
	u32			ix;
	segment*	seg;
	unspos		prevPos2;
	sgnpos		diag, prevDiag;

	sym1 = _sym1;
	sym2 = _sym2;

	prevDiag = 0;  prevPos2 = 0;
	for (ix=0,seg=st->seg ; ix<st->len ; ix++,seg++)
		{
		if (_sym1 == NULL) sym1 = "+";
		if (_sym2 == NULL) sym2 = ((seg->id & rcf_rev) != 0)? "-" : "+";

		diag = diagNumber (seg->pos1, seg->pos2);
#ifdef show_separation
		if ((ix != 0) && (diag == prevDiag))
			fprintf (f, "[" unsposFmt "] " unsposSlashSFmt " %d " scoreFmtSimple
			            "; id %d, diag " sgnposFmt " * (" unsposFmt "), tied_length=" possumFmt "\n",
						ix, seg->pos1+1, sym1, seg->pos2+1, sym2, seg->length,
						seg->s, seg->id, diag, seg->pos2-prevPos2, seg->scoreCov);
		else
#endif // show_separation
			fprintf (f, "[%d] " unsposSlashSFmt " " unsposFmt " " scoreFmtSimple
			            "; id %d, diag " sgnposFmt ", tied_length=" possumFmt "\n",
						ix, seg->pos1+1, sym1, seg->pos2+1, sym2, seg->length,
						seg->s, seg->id, diag, seg->scoreCov);
		prevDiag = diag;
		prevPos2 = seg->pos2;
		}

	}

//----------
//
// validate_heap--
//	Check whether a segment table is a valid min-heap, for debugging.
//
//----------
//
// Arguments:
//	segtable*	st:		The segment table to check.
//	char*		msg:	A message to output if validate fails.
//
// Returns:
//	(nothing;  failure causes program fatality)
//
//----------

#ifdef debugBinaryHeap

static void validate_heap
   (segtable*	st,
	char*		msg)
	{
	possum		scoreCov;
	segment*	seg, *lftChild, *rgtChild;
	int			ix, lftIx, rgtIx;

	if (st->coverage < st->coverageLimit)
		suicidef ("%s, below coverage limit", msg);

	for (ix=0,seg=st->seg ; ix<st->len ; ix++,seg++)
		{
		scoreCov = (possum) seg->length;

		lftIx = 2*ix+1;
		if (lftIx < st->len)
			{
			lftChild = &st->seg[lftIx];
			if (lftChild->s < seg->s)
				suicidef ("%s, node %d > node %d", msg, ix, lftIx);
			if (lftChild->s == seg->s)
				scoreCov += lftChild->scoreCov;

			rgtIx = lftIx + 1;
			if (rgtIx < st->len)
				{
				rgtChild = &st->seg[rgtIx];
				if (rgtChild->s < seg->s)
					suicidef ("%s, node %d > node %d", msg, ix, rgtIx);
				if (rgtChild->s == seg->s)
					scoreCov += rgtChild->scoreCov;
				}
			}

		if (scoreCov != seg->scoreCov)
			suicidef ("%s, node %d has bad score coverage", msg, ix);
		}

	}

#endif // debugBinaryHeap

