//-------+---------+---------+---------+---------+---------+---------+--------=
//
// File: sequences.h
//
//----------

#ifndef sequences_H				// (prevent multiple inclusion)
#define sequences_H

// other files

#include <stdio.h>				// standard C i/o stuff
#include "utilities.h"			// utility stuff
#include "dna_utilities.h"		// dna/scoring stuff

// establish ownership of global variables

#ifdef sequences_owner
#define global
#else
#define global extern
#endif

// "deep link" control variable access

#ifdef sequences_owner
int sequences_keepFastaArrow = false;	// true => keep ">" on header in
										// .. load_fasta_sequence()
#else
global int sequences_keepFastaArrow;
#endif

//----------
//
// sequence sizing types
//	These types control the range of sequence lengths we can handle before some
//	values will overflow.  For the sake of efficiency, we generally do not check
//	for overflow when indexing sequences.
//
//	Sequence lengths are normally assumed to be small enough to fit into a
//	31-bit integer.  This gives a maximum length of about 2.1 billion bp, which
//	is half the length of a (hypothetical) monoploid human genome.  The
//	programmer can override this at compile time by defining max_sequence_index
//	as 32 or 63.
//
//	maxSequenceLen (L) is the longest sequence that we intend to support.
//
//	unspos is an unsigned type that can represent 0..L.
//
//	sgnpos is a signed type that can represent -L..L.
//
//	possum is an unsigned type that can represent a sum of values in the 0..L
//	range.
//
//----------

#define maxSequenceOverrun 10

#if defined(max_sequence_index)
#define maxSequenceIndex max_sequence_index
#else
#define maxSequenceIndex 31
#endif

#if (maxSequenceIndex == 31)
#define unspos_sz 4
typedef u32 unspos;
typedef s32 sgnpos;
typedef u64 possum;
#elif (maxSequenceIndex == 32)
#define unspos_sz 4
typedef u32 unspos;
typedef s64 sgnpos;
typedef u64 possum;
#elif (maxSequenceIndex == 63)
#define unspos_sz 8
typedef u64 unspos;
typedef s64 sgnpos;
typedef u64 possum;
#else
#error ***** undecipherable max sequence length definition *****
#endif

#define maxSequenceLen ((unspos)(((1ULL)<<maxSequenceIndex)-(maxSequenceOverrun+1)))
#define seqposInfinity ((unspos) -1LL)

#if (maxSequenceIndex == 31)
#define unsposFmt     "%u"
#define unsposStarFmt "%*u"
#define sgnposFmt     "%d"
#define possumFmt     "%ju"
#elif (maxSequenceIndex == 32)
#define unsposFmt     "%u"
#define unsposStarFmt "%*u"
#define sgnposFmt     "%jd"
#define possumFmt     "%ju"
#elif (maxSequenceIndex == 63)
#define unsposFmt     "%ju"
#define unsposStarFmt "%*ju"
#define sgnposFmt     "%jd"
#define possumFmt     "%ju"
#endif

#define unsposSlashFmt  unsposFmt "/"   unsposFmt
#define unsposSlashSFmt unsposFmt "%s/" unsposFmt "%s"
#define unsposSlashCFmt unsposFmt "%c/" unsposFmt "%c"
#define unsposCommaFmt  unsposFmt ","   unsposFmt
#define unsposDashFmt   unsposFmt "-"   unsposFmt
#define unsposDotsFmt   unsposFmt ".."  unsposFmt

//----------
//
// sequence data structures and types
//
//----------
//
// sequences--
//	A sequence is a vector of characters (unsigned 8-bit values).
//
//----------
//
// Notes:
//	(1)	In the struct defintions, an (H) indicates pointers to items allocated
//		separately, and which therefore must be deallocated when the sequence is
//		deallocated.  (I) indicates pointers to items within the same structure.
//		(X) indicates pointers to other things but which do not need to be
//		deallocated.  (F) indicates files which must be closed when the
//		sequence is deallocated.
//
//	(2)	For partitioned sequences, p[0].sepPos is always 0 and p[len].sepPos is
//		always seq->len.  seq->v[p[ix].sepPos] is a NUL, for all ix (even for 0
//		and len).  All partitions are bounded on both sides by a NUL, even the
//		first partition.
//
//----------

#define seqBufferSize   128
#define maxSequenceName 100

typedef struct partition
	{
	unspos	sepPos;				//     the position (in seq->v) of the
								//     .. separating NUL preceding this
								//     .. partition;  valid only for entries
								//     .. 0..len in seqpartition->p;  for
								//     .. entry seqpartition.len, it is the
								//     .. position of the final NUL in seq->v;
								//     .. see note (2) above
	int		contig;				//     the contig number of this partition
								//     .. in the actual sequence file;  valid
								//     .. only for entries 0..len-1 in
								//     .. seqpartition->p
	char*	header;				// (X) the header of this partition;  valid
								//     .. only for entries 0..len-1 in
								//     .. seqpartition->p;  points into
								//     .. seqpartition->pool
	} partition;


typedef struct seqpartition
	{
	// information for a partitioned sequence-- a single sequence which holds
	// several actual sequences, separated by a NUL character (a zero).
	
	int		state;				//     the current initialization state;  one
								//     .. of seqpart_xxx below

	u32		size;				//     the number of entries allocated for
								//     .. header[], contig[], and sepPos[];
								//     .. note that this must always be at
								//     .. least 1 more than len
	u32		len;				//     the number of partitions of the sequence

	partition*	p;				// (H) an array, indexed by 0..len, of the
								//     .. partitions

	u32		poolSize;			//     the number of bytes allocated (size) and
	u32		poolLen;			//     .. actually used (len) for pool[]
	char*	pool;				// (H) an array containing the characters for
								//     .. header[]
	} seqpartition;

enum							// states for a partitioned sequence (only valid
	{							// .. if seqpartition.p != NULL
	seqpart_empty = 0,			// created by no partitions loaded yet
	seqpart_loading,			// some partitions loaded, but not all
	seqpart_ready				// all partitions loaded
	};

typedef struct twobit
	{
	// information for 2bit files only

	int		bigEndian;			//     true  => file was written as big endian
								//     false => written as little endian
	int		numContigs;			//     the number of contigs in the file (this
								//     .. is the sequenceCount field)
	long int indexFilePos;		//     position (as per ftell) of the beginning
								//     .. of the file index
	long int contigFilePos;		//     position (as per ftell) of the next entry
								//     .. in the file index

	char*	contigOfInterest;	// (H) the name of the only sequence of
								//     .. interest;  NULL means we're interested
								//     .. in every sequence
	int		contigLoaded;		//     true => the sequence corresponding to
								//             .. seq.contig has been loaded
								//             .. into v[]

	u32		nBlocksSize;		//     "temporary" arrays to hold n-blocks;
	u32*	nBlockStarts;		// (H) .. nBlocksSize is the number of entries
	u32*	nBlockSizes;		// (H) .. *allocated* for nBlockStarts[] and
								//     .. nBlockSizes[]

	u32		mBlocksSize;		//     "temporary" arrays to hold mask-blocks;
	u32*	mBlockstarts;		// (H) .. mBlocksSize is the number of entries
	u32*	mBlocksizes;		// (H) .. *allocated* for mBlockstarts[] and
								//     .. mBlocksizes[]
	} twobit;


typedef struct seq
	{
	// sequence content

	unspos	size;				//     the number of bytes allocated for v[]
	unspos	len;				//     the number of characters in the sequence,
								//     .. not including a terminating zero
	u8*		v;					// (H) the sequence content;  if v is not NULL,
								//     .. v[len] is a terminating zero
	unspos	start;				//     1-based starting location;  this is only
								//     .. valid after a sequence is loaded
	int		revCompFlags;		//     two bits describing how this sequence
								//     .. relates to what was read from the
								//     .. file; the four values are the rcf_xxx
								//     .. values defined below
	int		contig;				//     the number of the subsequence of the
								//     .. actual file (provided for fasta and
								//     .. 2bit files);  the first contig is 1
	int		preLoaded;			//     true => the data currently in v[] has
								//             .. been pre-loaded from the file
								//             .. behind the caller's back

	int		lockedHeader;		//     true => don't change header or
								//             .. shortHeader
	u32		headerSize;			//     number of bytes allocated for header[]
	char*	header;				// (H) the sequence's header (e.g. for each
								//     .. sequence in a fasta file);  this may
								//     .. be NULL
	u32		shortHeaderSize;	//     number of bytes allocated for
								//     .. shortHeader[]
	char*	shortHeader;		// (H) short version of the header;  this may be
								//     .. NULL

	// file containing the sequence

	char*	fileName;			// (H) the name of the file associated with this
								//     .. sequence;  this may be NULL
	FILE*	f;					// (F) the file associated with this sequence;
								//     .. this can be NULL (for which fileType
								//     .. must be seq_type_nofile)
	int		fileType;			//     the type of file being read (one of
								//     .. seq_type_xxx)
	int		rewindable;			//     true  => we can rewind the file
								//     false => we cannot rewind the file
								//     -1    => we do not know if we can rewind
								//              .. the file or not

	u32		pendingLen;			//     characters that have been read from the
	char*	pendingChars;		// (H) .. file but which have not been consumed;
	char*	pendingStack;		// (X) .. pendingChars is an array of size
			    				//     .. seqBufferSize;  pendingLen gives the
			    				//     .. number of non-consumed characters
								//     .. remaining;  pendingStack points to the
								//     .. top of the stack, which builds down
								//     .. from pendingChars+seqBufferSize

	twobit	twoBit;				//     additional info for 2bit files (only
								//     .. valid if fileType == seq_type_2bit)
	seqpartition partition;		//     additional info for partitioned
								//     .. sequences (only valid if partition.p
								//     .. is not NULL)

	// saved file state

	int		hasSavedState;		//     true => savedFilePos is meaningful
	long int savedFilePos;		//     (as per ftell)

	// sequence read control options

	char*	namesFileName;		// (H) the name of a file containing a list of
								//     .. contig names;  this may be NULL
	FILE*	namesFile;			// (F) file corresponding to namesFileName
	char	nextContigName	 	//     the name of the next contig-of-interest;
			    [maxSequenceName+1];// .. only valid if namesFile is not NULL
	int		contigPending;		//     true => the sequence file has already
								//             .. been scanned to locate the
								//             .. sequence for nextContigName,
								//             .. is positioned appropriately,
								//             .. and that sequence hasn't been
								//             .. loaded yet;  only valid if
								//             .. namesFile is not NULL

	char*	maskFileName;		// (H) the name of a file containing masking
								//     .. info to apply;  this may be NULL
	unspos	startLimit,			//     1-based starting and ending locations
			endLimit;			//     .. (inclusive) limiting the part of the
								//     .. sequence we are to read;  these are
								//     .. zero if there are no limits

	int		doRevComp;			//     true  => any sequence we read is to be
								//              .. reverse complemented
	int		doUnmask;			//     true  => any sequence we read is to be
								//              .. unmasked (converted to upper
								//              .. case)
	int		doJoin;				//     true  => combine the file's sequences
								//              .. into a partitoned sequence
	int		useFullNames;		//     true  => report full names in alignments
								//     false => report short names instead
	int		nameParseType;		//     how to parse sequecne headers, it at all;
								//     .. one of name_parse_type_xxx
	char*	nameTrigger;		//     a string to trigger the fetch of the
								//     .. short sequence name from a fasta
								//     .. header line;  e.g. "name:" means to
								//     .. fetch the name stating after the ":";
								//     .. NULL indicates an absent trigger
	int		prohibitAmbiDNA;	//     (this applies only to reading fasta files)
								//     false => permit ambiguous DNA characters
								//              .. B,D,H,K,M,R,S,V,W,Y
								//     true  => only A,C,G,T,N,X permitted
	} seq;


enum
	{
	seq_type_nofile=0,
	seq_type_unknown,
	seq_type_fasta,				// dna, permitting abiguity
	seq_type_nib,				// nybble-coded dna
	seq_type_2bit,				// 2bit-coded dna
	seq_type_qdna				// quantum-dna (byte-coded prob'listic ambiguity)
	};

enum
	{
	rcf_forward = 0,			// sequence is in normal order, uncomplemented
	rcf_comp    = 1,			// sequence is complemented but not reversed
	rcf_rev     = 2,			// sequence is reversed but not complemented
	rcf_revcomp = 3				// sequence is reversed and complemented
	};

enum
	{
	name_parse_type_core=0,		// trim path on left, junk and ext on right
	name_parse_type_alnum,		// trim path on left, use only alphanumeric and
								// .. underscore
	name_parse_type_trigger		// use nameTrigger
	};


//----------
//
// prototypes for routines in sequences.c
//
//----------

seq*  open_sequence_file   (char* name, int fileType);
seq*  open_rewindable_sequence_file (char* name, int fileType);
void  rewind_sequence_file (seq* seq);
seq*  copy_sequence        (seq* seq);
seq*  new_sequence         (unspos allocLen);
void  sequence_long_enough (seq* seq, unspos allocLen, int anticipate);
void  free_sequence        (seq* seq);
int   load_sequence        (seq* seq);
int   another_sequence     (seq* seq);
partition* lookup_partition (seq* seq, unspos pos);
char* species_and_chrom    (char* seqFileName);
void  print_sequence       (FILE* f, seq* seq, char* header, int perLine);
void  mask_sequence        (seq* seq, char* maskFileName);
void  revcomp_sequence     (seq* seq);
void  revcomp_in_place     (u8* seq, unspos seqLen);
void  backward_sequence    (seq* seq);
void  backward_in_place    (u8* seq, unspos seqLen);
void  upper_sequence       (seq* seq);
void  upper_in_place       (u8* seq, unspos seqLen);
void  match_composition    (seq* seq1, unspos pos1,
                            seq* seq2, unspos pos2, unspos length,
                            unspos count[4][4]);
int   percent_identical    (seq* seq1, unspos pos1,
                            seq* seq2, unspos pos2, unspos length);
score score_match          (scoreset* scoring,
                            seq* seq1, unspos pos1,
                            seq* seq2, unspos pos2, unspos length);
void  dump_aligned_nucleotides (FILE* f,
                            seq* seq1, unspos pos1,
                            seq* seq2, unspos pos2, unspos length);

#undef global
#endif // sequences_H
